from typing import Optional, List, Dict, Any
import numpy as np

from .. import World
from ..utils import logging, misc


def default_terminal_condition(task, agent_id, **kwargs):
    """ An example definition of terminal condition. """

    agent = [_a for _a in task.world.agents if _a.id == agent_id][0]

    def _check_out_of_lane():
        road_half_width = agent.trace.road_width / 2.
        return np.abs(agent.relative_state.x) > road_half_width

    def _check_exceed_max_rot():
        maximal_rotation = np.pi / 10.
        return np.abs(agent.relative_state.theta) > maximal_rotation

    out_of_lane = _check_out_of_lane()
    exceed_max_rot = _check_exceed_max_rot()
    done = out_of_lane or exceed_max_rot or agent.done
    other_info = {
        'done': done,
        'out_of_lane': out_of_lane,
        'exceed_max_rot': exceed_max_rot,
    }

    return done, other_info


def default_reward_fn(task, agent_id, **kwargs):
    """ An example definition of reward function. """
    reward = -1 if kwargs['done'] else 0  # simply encourage survival

    return reward, {}


class LaneFollowing:
    """ This class defines a simple lane following task in Vista. It basically
    handles vehicle state update of the ego car, rendering of specified sensors,
    and determing reward and terminal condition. The default terminal condition
    is set to (1) being out of lane (2) exceed maximal roation (3) reaching the
    end of the trace.

    Args:
        trace_paths (List[str]): A list of trace paths.
        trace_config (Dict): Configuration of the trace.
        car_configs (List[Dict]): Configuration of ``every`` cars.
        sensors_configs (List[Dict]): Configuration of sensors on ``every`` cars.
        task_config (Dict): Configuration of the task, which specifies reward function
                            and terminal condition. For more details, please check the
                            source code.
        logging_level (str): Logging level (``DEBUG``, ``INFO``, ``WARNING``,
                             ``ERROR``, ``CRITICAL``); default set to ``WARNING``.

    """
    DEFAULT_CONFIG = {
        'reward_fn': default_reward_fn,
        'terminal_condition': default_terminal_condition,
    }

    def __init__(self,
                 trace_paths: List[str],
                 trace_config: Dict,
                 car_config: Dict,
                 sensors_configs: Optional[List[Dict]] = [],
                 task_config: Optional[Dict] = dict(),
                 logging_level: Optional[str] = 'WARNING'):
        logging.setLevel(getattr(logging, logging_level))

        self._config = misc.merge_dict(task_config, self.DEFAULT_CONFIG)

        self._world: World = World(trace_paths, trace_config)
        agent = self._world.spawn_agent(car_config)
        for sensor_config in sensors_configs:
            sensor_type = sensor_config.pop('type')
            if sensor_type == 'camera':
                agent.spawn_camera(sensor_config)
            elif sensor_type == 'event_camera':
                agent.spawn_event_camera(sensor_config)
            elif sensor_type == 'lidar':
                agent.spawn_lidar(sensor_config)
            else:
                raise NotImplementedError(
                    f'Unrecognized sensor type {sensor_type}')

        self._distance = 0
        self._prev_xy = np.zeros((2, ))

    def reset(self):
        """ Reset the environment. This basically reset the :class:`World` object
        in Vista, which reset all (the only) agent in the world.

        Returns:
            Dict: A dictionary with keys as agent IDs and values as observation
            for each agent, which is also a dictionary with keys as sensor IDs
            and values as sensory measurements.

        """
        self._world.reset()
        agent = self._world.agents[0]
        observations = self._append_agent_id(agent.observations)
        self._distance = 0
        self._prev_xy = np.zeros((2, ))
        return observations

    def step(self, action, dt=1 / 30.):
        """ Step the environment. This involves updating agent's state based on
        the given actions and determining reward and termination.

        Args:
            actions (Dict[str, np.ndarray]):
                A dictionary with keys as agent IDs and values as actions
                to be executed to interact with the environment and other
                agents.
            dt (float): Elapsed time in second; default set to 1/30.

        Returns:
            Return a tuple (``dict_a``, ``dict_b``, ``dict_c``, ``dict_d``),
            where ``dict_a`` is the observation, ``dict_b`` is the reward,
            ``dict_c`` is whether the episode terminates, ``dict_d`` is additional
            informations for every agents; keys of every dictionary are agent IDs.

        """
        # Step agent and get observation
        agent = self._world.agents[0]
        action = np.array([action[agent.id][0], agent.human_speed])
        agent.step_dynamics(action, dt=dt)
        agent.step_sensors()
        observations = agent.observations

        # Define terminal condition
        done, info_from_terminal_condition = self.config['terminal_condition'](
            self, agent.id)

        # Define reward
        reward, _ = self.config['reward_fn'](self, agent.id,
                                             **info_from_terminal_condition)

        # Get info
        info = misc.fetch_agent_info(agent)
        info['out_of_lane'] = info_from_terminal_condition['out_of_lane']
        info['exceed_rot'] = info_from_terminal_condition['exceed_rot']

        current_xy = agent.ego_dynamics.numpy()[:2]
        self._distance += np.linalg.norm(current_xy - self._prev_xy)
        self._prev_xy = current_xy
        info['distance'] = self._distance

        # Pack output
        observations, reward, done, info = map(
            self._append_agent_id, [observations, reward, done, info])

        return observations, reward, done, info

    def set_seed(self, seed) -> None:
        """ Set random seed.

        Args:
            seed (int): Random seed.

        """
        self._seed = seed
        self._rng = np.random.default_rng(self.seed)
        self.world.set_seed(seed)

    def _append_agent_id(self, data):
        agent = self._world.agents[0]
        return {agent.id: data}

    @property
    def config(self) -> Dict:
        """ Configuration of this task. """
        return self._config

    @property
    def world(self) -> World:
        """ :class:`World` of this task. """
        return self._world

    @property
    def seed(self) -> int:
        """ Random seed for the task and the associated :class:`World`. """
        return self._seed



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [ ]

setup_requirements = [ ]

test_requirements = [ ]

setup(
    author="Alexander Amini",
    author_email='amini@mit.edu',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Data-driven simulation for training and evaluating full-scale autonomous vehicles",
    long_description=readme,
    install_requires=requirements,
    license="MIT license",
    include_package_data=True,
    keywords='vista',
    name='vista',
    packages=find_packages(include=['vista', 'vista.*']),
    setup_requires=setup_requirements,
    url='https://github.com/vista-simulator/vista',
    download_url = 'https://github.com/vista-simulator/vista/archive/v2.0.3.tar.gz',
    version='2.0.3',
    zip_safe=False,
)



""" Helper functions for spatial transformation. Follow right handed
    (OpenGL) coordinate system. """
from typing import Optional, Union, List, Tuple, Any
import numpy as np
from scipy.spatial.transform import Rotation

Vec = Union[np.ndarray, List[Any], Tuple[Any]]


def rot2mat(rot: Vec, seq: Optional[str] = 'xyz') -> np.ndarray:
    """ Convert euler vector (with sequence order) to rotation matrix.

    Args:
        rot (Vec): A 3-dimensional rotation vector in Euler angle.
        seq (str): The order of the rotation vector, e.g., ``xzy``;
                   default to ``xyz``.

    Returns:
        np.ndarray: A rotation matrix equivalent to the given rotation vector.

    """
    R = Rotation.from_euler(seq, rot)
    return R.as_matrix()


def vec2mat(trans: Vec, rot: Vec) -> np.ndarray:
    """ Convert translation and rotation vector to transformation matrix.

    Args:
        trans (Vec): A 3-dimensional translation vector.
        rot (Vec): A 3-dimensional rotation vector.

    Returns:
        np.ndarray: A 4-by-4 transformation matrix in SE(3).

    """
    mat = np.eye(4)
    mat[:3, 3] = trans
    mat[:3, :3] = rot2mat(rot)
    return mat


def euler2quat(euler: Vec,
               seq: Optional[str] = 'xyz',
               degrees: Optional[bool] = False) -> Vec:
    """ Convert Euler rotation to quaternion.

    Args:
        euler (Vec): A 3-dimensional rotation vector in Euler angle.
        seq (str): The order of the rotation vector, e.g., ``xzy``;
                   default to ``xyz``.

    Returns:
        Vec: A 4-dimensional vector that describes a quaternion.

    """
    R = Rotation.from_euler(seq, euler, degrees)
    return R.as_quat()


def quat2euler(quat: Vec,
               seq: Optional[str] = 'xyz',
               degrees: Optional[bool] = False) -> Vec:
    """ Convert quaternion to Euler rotation.

    Args:
        quat (Vec): A 4-dimensional vector that describes a quaternion.
        seq (str): The order of the output Euler rotation vector; default
                   to ``xyz``.
        degrees (bool): Whether to convert the output rotation to degrees
                        instead of radians; default to ``False``.

    Returns:
        Vec: A 3-dimensional rotation vector in Euler angle.

    """
    R = Rotation.from_quat(quat)
    return R.as_euler(seq, degrees)


def latlongyaw2vec(latlongyaw: Vec) -> Tuple[Vec, Vec]:
    """ Convert lateral, longitudinal, yaw compoenents to translational and
    rotational vectors. Note that the longitudinal component will be negated
    to follow right-handed OpenGL coordinate system.

    Args:
        latlongyaw (Vec): A 3-dimensional vector with entries as lateral shift,
                          longitudinal shift, and yaw difference.

    Returns:
        Return a tuple (``Vec_a``, ``Vec_b``), where ``Vec_a`` is a 3-dimensional
        translation vector and ``Vec_b`` is a 3-dimensional rotation vector in
        Euler angle ``xyz``.

    """
    lat, long, yaw = latlongyaw
    trans = np.array([lat, 0., -long])
    rot = np.array([0., yaw, 0.])
    return trans, rot


def vec2latlongyaw(trans: Vec, rot: Vec) -> Vec:
    """ Convert translational and rotational vectors to lateral,
    longitudinal, yaw compoenents. Note that the y translational
    component will be negated to follow right-handed OpenGL
    coordinate system.

    Args:
        trans (Vec): A 3-dimensional translation vector.
        rot (Vec): A 3-dimensional Euler rotation vector.

    Returns:
        Vec: A 3-dimensional vector with entries as lateral shift,
             longitudinal shift, and yaw difference.

    """
    return np.array([trans[0], -trans[2], rot[1]])


def compute_relative_latlongyaw(latlongyaw: Vec, latlongyaw_ref: Vec) -> Vec:
    """ Compute relative lateral, longitudinal, yaw compoenents.

    Args:
        latlongyaw (Vec): A 3-dimensional vector with lateral,
                          longitudinal, and yaw component.
        latlongyaw_ref (Vec): A reference 3-dimensional vector with lateral,
                              longitudinal, and yaw component.

    Returns:
        Vec: A 3-dimensional vector that describe relative lateral,
             longitudinal shift and yaw difference between the given
             two vectors.

    """
    mat = vec2mat(*latlongyaw2vec(latlongyaw))
    mat_ref = vec2mat(*latlongyaw2vec(latlongyaw_ref))
    rel_mat = np.matmul(SE3_inv(mat_ref), mat)
    rel_trans, rel_rot = mat2vec(rel_mat)
    rel_xyyaw = vec2latlongyaw(rel_trans, rel_rot)
    return rel_xyyaw


def SE3_inv(T_in: np.ndarray) -> np.ndarray:
    """ More efficient matrix inversion for SE(3).

    Args:
        T_in (np.ndarray): a 4-by-4 transformation matrix in SE(3).

    Returns:
        np.ndarray: The inverse of ``T_in``.

    """
    R_in = T_in[:3, :3]
    t_in = T_in[:3, [-1]]
    R_out = R_in.T
    t_out = -np.matmul(R_out, t_in)
    return np.vstack((np.hstack((R_out, t_out)), np.array([0, 0, 0, 1])))


def mat2vec(mat: np.ndarray,
            seq: Optional[str] = 'xyz',
            degrees: Optional[bool] = False) -> Tuple[Vec, Vec]:
    """ Convert transformation matrix to translational and rotational vectors.

    Args:
        mat (np.ndarray): A 4-by-4 transformation matrix in SE(3).
        seq (str): The order of the output Euler rotation vector; default
                   to ``xyz``.
        degrees (bool): Whether to convert the output rotation to degrees
                        instead of radians; default to ``False``.

    Returns:
        Return a tuple (``Vec_a``, ``Vec_b``), where ``Vec_a`` is a 3-dimensional
        translation vector and ``Vec_b`` is a 3-dimensional rotation vector in
        Euler angle ``xyz``.

    """
    trans = mat[:3, 3]
    R = Rotation.from_matrix(mat[:3, :3])
    rot = R.as_euler(seq, degrees)
    return trans, rot


def pi2pi(angle: float) -> float:
    """ Make sure angle is within -pi to pi.

    Args:
        angle (float): Input angle in radians.

    Returns:
        float: Output angle that is within range [-pi, pi]

    """
    if angle >= np.pi:
        angle -= 2 * np.pi
    if angle <= -np.pi:
        angle += 2 * np.pi
    return angle



import os
import sys
from typing import Dict, List, Optional
import numpy as np
from ffio import FFReader
import cv2
import torch

from .camera_utils import CameraParams, ViewSynthesis
from .BaseSensor import BaseSensor
from .Camera import Camera
from ..Entity import Entity
from ...utils import logging, misc

try:
    from metavision_core.event_io.raw_reader import RawReader
except ImportError:
    logging.warning(
        'Fail to import module for event camera. Remember to do ' +
        'source <some-dir>/openeb/build/utils/scripts/setup_env.sh' +
        'Can ignore this if not using it')


class EventCamera(BaseSensor):
    """ A event camera sensor object that synthesizes event data locally around the RGB
    dataset given a viewpoint (potentially different from the dataset) and timestamp with
    video interpolation and an event emission model.

    Args:
        attach_to (Entity): A parent object (car) to be attached to.
        config (Dict): Configuration of the sensor. An example (default) config is,

            >>> DEFAULT_CONFIG = {
                'rig_path': None,
                # Event camera
                'name': 'event_camera_front',
                'original_size': (480, 640),
                'size': (240, 320),
                'optical_flow_root': '../data_prep/Super-SloMo',
                'checkpoint': '../data_prep/Super-SloMo/ckpt/SuperSloMo.ckpt',
                'lambda_flow': 0.5,
                'max_sf': 16,
                'use_gpu': True,
                'positive_threshold': 0.1,
                'sigma_positive_threshold': 0.02,
                'negative_threshold': -0.1,
                'sigma_negative_threshold': 0.02,
                'reproject_pixel': False,
                'subsampling_ratio': 0.5,
                # RGB rendering
                'base_camera_name': 'camera_front',
                'base_size': (600, 960),
                'depth_mode': 'FIXED_PLANE',
                'use_lighting': False,
            }

    Note that event camera simulation requires third-party dependence and pretrained
    checkpoint for video interpolation.

    """
    DEFAULT_CONFIG = {
        'rig_path': None,
        # Event camera
        'name': 'event_camera_front',
        'original_size': (480, 640),
        'size': (240, 320),
        'optical_flow_root': '../data_prep/Super-SloMo',
        'checkpoint': '../data_prep/Super-SloMo/ckpt/SuperSloMo.ckpt',
        'lambda_flow': 0.5,
        'max_sf': 16,
        'use_gpu': True,
        'positive_threshold': 0.1,
        'sigma_positive_threshold': 0.02,
        'negative_threshold': -0.1,
        'sigma_negative_threshold': 0.02,
        'reproject_pixel': False,
        'subsampling_ratio': 0.5,
        # RGB rendering
        'base_camera_name': 'camera_front',
        'base_size': (600, 960),
        'depth_mode': 'FIXED_PLANE',
        'use_lighting': False,
    }

    def __init__(self, attach_to: Entity, config: Dict) -> None:
        super(EventCamera, self).__init__(attach_to, config)

        self._config['rig_path'] = os.path.expanduser(self._config['rig_path'])
        self._camera_param: CameraParams = CameraParams(
            self._config['rig_path'], self.name)
        self._camera_param.resize(*self._config['size'])
        self._base_camera_param: CameraParams = CameraParams(
            self._config['rig_path'], self._config['base_camera_name'])
        self._base_camera_param.resize(*self._config['base_size'])
        self._config['use_synthesizer'] = self._config.get(
            'use_synthesizer', True)
        if self._config['use_synthesizer']:
            self._streams: Dict[str, FFReader] = dict()
            vs_cam_param = self._base_camera_param if \
                self.config['reproject_pixel'] else self._camera_param
            self._view_synthesis: ViewSynthesis = ViewSynthesis(
                vs_cam_param, self._config, init_with_bg_mesh=False)

            # load video interpolation model
            sys.path.append(os.path.abspath(self._config['optical_flow_root']))
            from slowmo_warp import SlowMoWarp
            self._config['checkpoint'] = os.path.abspath(
                self._config['checkpoint'])
            height = self._config['base_size'][0] if \
                self.config['reproject_pixel'] else self._config['size'][0]
            width = self._config['base_size'][1] if \
                self.config['reproject_pixel'] else self._config['size'][1]
            self._interp = SlowMoWarp(height=height,
                                      width=width,
                                      checkpoint=self._config['checkpoint'],
                                      lambda_flow=self._config['lambda_flow'],
                                      cuda=self._config['use_gpu'])
            self._prev_frame: np.ndarray = None
            self._prev_timestamp: float = None
        else:
            self._streams: Dict[str, RawReader] = dict()
            self._prev_timestamp: float = None

    def reset(self) -> None:
        """ Reset Event camera sensor by initiating RGB data stream (as it's simulated
        using RGB data) based on current reference pointer to the dataset.

        """
        logging.info(f'Event camera ({self.id}) reset')

        if self._config['use_synthesizer']:
            # Close stream already open
            for stream in self.streams.values():
                stream.close()

            # Fetch video stream from the associated trace. This is mostly the same as how we handle
            # video stream in Camera object. The difference and the reason of not sharing with
            # Camera is (1) EventCamera can be a standalone object to be initiated and more
            # importantly (2) TODO
            multi_sensor = self.parent.trace.multi_sensor
            if self.name == multi_sensor.main_event_camera:
                for camera_name in multi_sensor.camera_names:  # base cameras are regular cameras
                    # get video stream
                    video_path = os.path.join(self.parent.trace.trace_path,
                                              camera_name + '.avi')
                    cam_h, cam_w = self.base_camera_param.get_height(
                    ), self.base_camera_param.get_width()
                    stream = FFReader(video_path,
                                      custom_size=(cam_h, cam_w),
                                      verbose=False)
                    self._streams[camera_name] = stream

                    # seek based on timestamp
                    frame_num = self.parent.trace.good_frames[camera_name][
                        self.parent.segment_index][self.parent.frame_index]
                    seek_sec = self._streams[camera_name].frame_to_secs(
                        frame_num)
                    self._streams[camera_name].seek(seek_sec)
            else:  # use shared streams from the main camera
                main_name = multi_sensor.main_event_camera
                main_sensor = [
                    _s for _s in self.parent.sensors if _s.name == main_name
                ]
                assert len(
                    main_sensor) == 1, 'Cannot find main sensor {}'.format(
                        main_name)
                main_sensor = main_sensor[0]
                assert isinstance(main_sensor,
                                  Camera), 'Main sensor is not Camera object'
                self._streams = main_sensor.streams

            # Add background mesh for all video stream in view synthesis. Note that we use camera
            # parameters from the base camera since 2D-to-3D part is determined by the regular RGB
            # video streams. Reprojecting back to 2D is then be handled by event camera
            # configuration
            parent_sensor_dict = {_s.name: _s for _s in self.parent.sensors}
            for camera_name in self.streams.keys():
                if camera_name not in self.view_synthesis.bg_mesh_names:
                    if camera_name in parent_sensor_dict.keys():
                        camera_param = parent_sensor_dict[
                            camera_name].camera_param
                    else:
                        camera_param = CameraParams(self._config['rig_path'],
                                                    camera_name)
                        camera_param.resize(*self._config['base_size'])
                    self.view_synthesis.add_bg_mesh(camera_param)

            # Reset previous frame
            self._prev_frame = None

            if self.config['reproject_pixel']:
                cam_h = self.camera_param.get_height()
                cam_w = self.camera_param.get_width()
                world_rays = self.view_synthesis._world_rays[
                    self.base_camera_param.name]
                K = self.camera_param.get_K()
                self._uv = np.matmul(K, world_rays).T[:, :2][:, ::-1]
        else:
            # TODO: check if we need to close event data stream
            logging.debug(
                'Not sure if it is ok to not close the event data stream')

            raw_path = os.path.join(self.parent.trace.trace_path,
                                    f'{self.name}.raw')
            stream = RawReader(raw_path)
            self._streams[self.name] = stream

            # Reset previous timestamp
            self._prev_timestamp = None

    def capture(self,
                timestamp: float,
                update_rgb_frame_only: Optional[bool] = False) -> np.ndarray:
        """ Synthesize event data based on current timestamp and transformation
        between the novel viewpoint to be simulated and the nominal viewpoint from
        the pre-collected RGB dataset. In a very high level, it basically performs
        video interpolation across consecutive RGB frames, extract events with an
        event emission model, projects the events to the event camera space. Note
        that the simulation is running a deep network (SuperSloMo) for video interpolation.

        Args:
            timestamp (float): Timestamp that allows to retrieve a pointer to
                the dataset for data-driven simulation (synthesizing RGB image
                from real RGB video).

        Returns:
            np.ndarray: A synthesized event data.

        """
        logging.info(f'Event camera ({self.id}) capture')

        if self._config['use_synthesizer']:
            # Get frame at the closest smaller timestamp from dataset. In event camera simulation,
            # we don't use flow interpolation since event generation is already continuous-time.
            multi_sensor = self.parent.trace.multi_sensor
            if self.name == multi_sensor.main_event_camera:
                all_frame_nums = multi_sensor.get_frames_from_times(
                    [timestamp], fetch_smaller=False)
                for camera_name in multi_sensor.camera_names:
                    # rgb camera stream
                    stream = self.streams[camera_name]
                    frame_num = all_frame_nums[camera_name][0]

                    if frame_num < stream.frame_num:
                        seek_sec = stream.frame_to_secs(frame_num)
                        stream.seek(seek_sec)

                    while stream.frame_num != frame_num:
                        stream.read()

            frames = dict()
            for camera_name in multi_sensor.camera_names:
                frames[camera_name] = self.streams[camera_name].image.copy()

            # Synthesis by rendering with compensation of relative pose between event camera
            # parameters and the base (RGB) camera parameters since we set the view synthesizer
            # with event camera parameters and project frame to 3D meshes using base camera
            # parameters
            lat, long, yaw = self.parent.relative_state.numpy()
            trans = np.array([lat, 0., -long])
            rot = np.array([0., yaw, 0.])
            rendered_frame, _ = self.view_synthesis.synthesize(
                trans, rot, frames)

            # work with RGB instead of BGR; need copy otherwise fail at converting to tensor
            rendered_frame = rendered_frame[:, :, ::-1].copy()

            if update_rgb_frame_only:
                self._prev_frame = rendered_frame
                self._prev_timestamp = timestamp
            else:
                events = [[], []]
                if self.prev_frame is not None:
                    with torch.no_grad():
                        out = self._interp.forward_warp(
                            self.prev_frame,
                            rendered_frame,
                            max_sf=self.config['max_sf'],
                            use_max_flow=True)

                    flow = list(
                        map(lambda _x: _x.cpu().numpy()[0].transpose(1, 2, 0),
                            out['flow']))

                    interp = [self.prev_frame] + out['interpolated']
                    if len(interp) >= 2:
                        last_interp_frame = cv2.cvtColor(
                            interp[0], cv2.COLOR_RGB2GRAY)
                        dt = (timestamp - self.prev_timestamp) / out['sf']
                        event_timestamp = self.prev_timestamp
                        for interp_frame in interp[1:]:
                            interp_frame = cv2.cvtColor(
                                interp_frame, cv2.COLOR_RGB2GRAY)
                            log_d_interp = np.log(interp_frame / 255. + 1e-8) - \
                                        np.log(last_interp_frame / 255. + 1e-8)
                            sample_fn = lambda mu, sig: np.clip(
                                np.random.normal(mu, sig), 0., 1.)
                            positive_C = sample_fn(
                                self.config['positive_threshold'],
                                self.config['sigma_positive_threshold'])
                            negative_C = -sample_fn(
                                -self.config['negative_threshold'],
                                self.config['sigma_negative_threshold'])
                            if self.config['reproject_pixel']:
                                cam_h = self.camera_param.get_height()
                                cam_w = self.camera_param.get_width()

                                def extract_uv(_mask):
                                    _fl_mask = _mask.flatten()
                                    _uv = self._uv[_fl_mask]

                                    _mode = 1
                                    if _mode == 0:
                                        _uv = np.round(_uv).astype(np.int)
                                    elif _mode == 1:
                                        _uv_floor = np.floor(_uv)
                                        _uv_ceil = np.ceil(_uv)
                                        diff_floor = np.abs(_uv - _uv_floor)
                                        diff_ceil = np.abs(_uv - _uv_ceil)

                                        n_uv = _uv.shape[0]
                                        zeros = np.zeros((n_uv, ))
                                        _uv_score = np.concatenate([
                                            np.stack([
                                                _uv_floor[:, 0],
                                                _uv_floor[:, 1], zeros
                                            ],
                                                     axis=1),
                                            np.stack([
                                                _uv_floor[:, 0],
                                                _uv_ceil[:, 1], zeros
                                            ],
                                                     axis=1),
                                            np.stack([
                                                _uv_ceil[:, 0],
                                                _uv_floor[:, 1], zeros
                                            ],
                                                     axis=1),
                                            np.stack([
                                                _uv_ceil[:, 0], _uv_ceil[:, 1],
                                                zeros
                                            ],
                                                     axis=1)
                                        ],
                                                                   axis=0)

                                        _uv_score[:n_uv,
                                                  2] = diff_ceil[:,
                                                                 0] * diff_ceil[:,
                                                                                1]
                                        _uv_score[
                                            n_uv:2 * n_uv,
                                            2] = diff_ceil[:,
                                                           0] * diff_floor[:,
                                                                           1]
                                        _uv_score[
                                            2 * n_uv:3 * n_uv,
                                            2] = diff_floor[:,
                                                            0] * diff_ceil[:,
                                                                           1]
                                        _uv_score[
                                            3 * n_uv:,
                                            2] = diff_floor[:,
                                                            0] * diff_floor[:,
                                                                            1]

                                        _uv = _uv_score[
                                            _uv_score[:, 2] >= 0.2, :2].astype(
                                                np.int)
                                    else:
                                        _uv_floor = np.floor(_uv)
                                        _uv_ceil = np.ceil(_uv)
                                        _uv = np.concatenate([
                                            np.stack([
                                                _uv_floor[:, 0], _uv_floor[:,
                                                                           1]
                                            ],
                                                     axis=1),
                                            np.stack([
                                                _uv_floor[:, 0], _uv_ceil[:, 1]
                                            ],
                                                     axis=1),
                                            np.stack([
                                                _uv_ceil[:, 0], _uv_floor[:, 1]
                                            ],
                                                     axis=1),
                                            np.stack([
                                                _uv_ceil[:, 0], _uv_ceil[:, 1]
                                            ],
                                                     axis=1)
                                        ],
                                                             axis=0)
                                        _uv = _uv.astype(np.int)

                                    _valid_mask = (_uv[:,0] > 0) & (_uv[:,0] < cam_h) & \
                                                (_uv[:,1] > 0) & (_uv[:,1] < cam_w)
                                    _uv = _uv[_valid_mask]
                                    return _uv

                                positive_uv = extract_uv(
                                    log_d_interp >= positive_C)
                                negative_uv = extract_uv(
                                    log_d_interp <= negative_C)
                            else:
                                positive_uv = np.argwhere(
                                    log_d_interp >= positive_C)
                                negative_uv = np.argwhere(
                                    log_d_interp <= negative_C)
                            event_timestamp += dt
                            event_timestamp_us = (event_timestamp *
                                                  1e6).astype(np.int64)
                            positive_events = np.concatenate([
                                positive_uv,
                                np.tile([event_timestamp_us, 1],
                                        (positive_uv.shape[0], 1))
                            ],
                                                             axis=1)
                            negative_events = np.concatenate([
                                negative_uv,
                                np.tile([event_timestamp_us, 1],
                                        (negative_uv.shape[0], 0))
                            ],
                                                             axis=1)
                            events[0].append(positive_events)
                            events[1].append(negative_events)
                            last_interp_frame = interp_frame

                self._prev_frame = rendered_frame
                self._prev_timestamp = timestamp

                return events
        else:
            if self._prev_timestamp is not None:
                stream = self.streams[self.name]

                multi_sensor = self.parent.trace.multi_sensor
                ref_sensor = multi_sensor.master_sensor
                # TODO: didn't do dedicated alignment between ros clock and event camera clock
                logging.debug(
                    'No dedicated alignment between ros and event camera clocks'
                )
                offset_timestamp = multi_sensor.get_time_from_frame_num(
                    ref_sensor, 0)

                event_timestamp_us = int(
                    (self._prev_timestamp - offset_timestamp) * 1e6)
                if event_timestamp_us < stream.current_time:
                    stream.reset()  # cannot seek backward
                stream.seek_time(event_timestamp_us)

                dt_us = int((timestamp - self._prev_timestamp) * 1e6)
                events = stream.load_delta_t(dt_us)
                if stream.is_done():
                    self.parent._done = True

                ori_cam_h, ori_cam_w = self.config['original_size']
                cam_h = self.camera_param.get_height()
                cam_w = self.camera_param.get_width()

                # NOTE: do subsampling otherwise too many events left
                subsample_factor = (cam_h * cam_w) / float(
                    ori_cam_h * ori_cam_w) * self.config['subsampling_ratio']
                n_events = events['x'].shape[0]
                subsample_mask = np.random.choice(np.arange(n_events),
                                                  size=int(n_events *
                                                           subsample_factor),
                                                  replace=False)
                events_x = events['x'][subsample_mask]
                events_y = events['y'][subsample_mask]
                events_p = events['p'][subsample_mask]

                mul_x = cam_w / float(ori_cam_w)
                mul_y = cam_h / float(ori_cam_h)
                events_x = (events_x * mul_x).astype(np.int)
                events_y = (events_y * mul_y).astype(np.int)

                positive_mask = events_p == 1
                negative_mask = events_p == 0
                positive_events = np.array(
                    [events_y[positive_mask], events_x[positive_mask]]).T
                negative_events = np.array(
                    [events_y[negative_mask], events_x[negative_mask]]).T

                events = [[positive_events], [negative_events]]
            else:
                events = [[], []]

            self._prev_timestamp = timestamp

            return events

    @property
    def config(self) -> Dict:
        """ Configuration of this sensor. """
        return self._config

    @property
    def streams(self) -> Dict[str, FFReader]:
        """ Data stream of RGB image/video dataset to be simulated from. """
        return self._streams

    @property
    def camera_param(self) -> CameraParams:
        """ Camera parameters of the virtual event camera. """
        return self._camera_param

    @property
    def base_camera_param(self) -> CameraParams:
        """ Camera parameters of the RGB camera. """
        return self._base_camera_param

    @property
    def view_synthesis(self) -> ViewSynthesis:
        """ View synthesizer object. """
        return self._view_synthesis

    @property
    def prev_frame(self) -> np.ndarray:
        """ Previous RGB frame. """
        return self._prev_frame

    @property
    def prev_timestamp(self) -> float:
        """ Previous timestamp. """
        return self._prev_timestamp



from typing import List, Dict, Tuple, Optional, Callable
import numpy as np

from .Trace import Trace
from ..entities.agents import Car
from ..utils import logging


class World:
    """ This class specifies the :class:`World` where all entities (i.e., agents and sensors) 
    in VISTA lives in. The is built upon one or multiple traces pre-collected from the real world. 
    Users can spawn agents (:class:`Car`) in the :class:`World`, where all agents will be visible 
    while rendering sensory measurement. At reset, the :class:`World` randomly sample an pointer
    in the dataset (conceptually retrieving a frame within a trace) as the initial condition for
    the simulation.

    Args:
        trace_paths (List(str)): A list of paths to traces
        trace_config (Dict): Configuration of traces

    Example usage::

        >>> world = vista.World(trace_paths, trace_config)
        >>> agent = world.spawn_agent(agent_config)
        >>> world.reset()

    """
    def __init__(
        self, trace_paths: List[str],
        trace_config: Optional[Dict] = dict()) -> None:
        # A list of traces that define the world
        self._traces: List[Trace] = [
            Trace(trace_path, trace_config) for trace_path in trace_paths
        ]

        # A list of agents within this world. Agents start in a single
        # trace but can be teleported between traces, since they are all in
        # the same world.
        self._agents: List[Car] = []

        # Random seed; will also assign the same seed to traces
        self.set_seed(0)

    def spawn_agent(self, config: Dict) -> Car:
        """ Spawn an agent in this world.

        Args:
            config (Dict): Configuration of the agent.

        Returns:
            Car: The agent being spawned.

        """
        agent = Car(world=self, car_config=config)
        logging.info('Spawn an agent ({})'.format(agent.id))
        self._agents.append(agent)

        return agent

    def reset(
        self, initial_dynamics_fn: Optional[Dict[str,
                                                 Callable]] = dict()) -> None:
        """ Reset the world. This includes (1) sample a new anchor point from the real-world
        dataset to be simulated from and (2) reset states for all agents.

        Args:
            initial_dynamics_fn (Dict[str, Callable]):
                A dict mapping agent names to a function that initialize agents poses.

        """
        logging.info('World reset')

        # Sample a new trace and a new location at the sampled trace
        new_trace_index, new_segment_index, new_frame_index = \
            self.sample_new_location()

        # Reset agents
        for agent in self.agents:
            agent.reset(new_trace_index, new_segment_index, new_frame_index,
                        initial_dynamics_fn.get(agent.id, None))

    def sample_new_location(self) -> Tuple[int, int, int]:
        """ Sample a pointer to the dataset for simulation.

        Returns:
            Return a tuple (``int_a``, ``int_b``, ``int_c``), where ``int_a`` is trace index,
            ``int_b`` is segment index, and ``int_c`` is frame index.

        """
        new_trace_index = self.sample_new_trace_index()
        trace = self.traces[new_trace_index]

        new_segment_index = trace.find_segment_reset()

        # note that this is the index of Trace.good_frames instead of element
        new_frame_index = trace.find_frame_reset(new_segment_index)

        return new_trace_index, new_segment_index, new_frame_index

    def sample_new_trace_index(self) -> int:
        """ Sample a new trace index based on number of frames in a trace.

        Returns:
            int: An index to specify which trace to be simulated from.

        """
        trace_reset_probs = np.zeros(len(self.traces))
        for i, trace in enumerate(self.traces):
            trace_reset_probs[i] = trace.num_of_frames
        trace_reset_probs /= np.sum(trace_reset_probs)

        new_trace_index = self._rng.choice(trace_reset_probs.shape[0],
                                           p=trace_reset_probs)

        return new_trace_index

    def set_seed(self, seed) -> None:
        """ Set random seed.

        Args:
            seed (int): Random seed.

        """
        self._seed = seed
        self._rng = np.random.default_rng(self.seed)
        [t.set_seed(seed) for t in self.traces]

    @property
    def seed(self) -> int:
        """ Random seed for sampling pointer to the dataset during :meth:`reset`. """
        return self._seed

    @property
    def traces(self) -> List[Trace]:
        """ All trace attached to the :class:`World`. """
        return self._traces

    @property
    def agents(self) -> List[Car]:
        """ All agents that live in the :class:`World`. """
        return self._agents

    def __repr__(self) -> str:
        return '<{}> '.format(self.__class__.__name__) + \
               '#traces: {} '.format(len(self.traces)) + \
               '#agents: {} '.format(len(self.agents))



import os
import re
import csv
from typing import Tuple
import numpy as np


class LabelSearch(object):
    """ This class handles annotations of the collected traces and filter out
    frames that are not good based on several fields including ``time of day``,
    ``weather``, ``road type``, ``maneuver``, ``direction``, ``tag``. Please
    check data annotation process for more details on each field.

    Args:
        time_of_day (str): Time of day to be considered as good frames.
        weather (str): Weather condition to be considered as good frames.
        road_type (str): Road type to be considered as good frames.
        maneuver (str): Human maneuver to be considered as good frames.
        direction (str): Direction to be considered as good frames.
        tag (str): Other tags to be considered as good frames.

    """
    FIELDS = [
        'timestamp', 'time_of_day', 'weather', 'road_type', 'maneuver',
        'direction', 'tag'
    ]

    def __init__(self, time_of_day: str, weather: str, road_type: str,
                 maneuver: str, direction: str, tag: str) -> None:
        self._time_of_day = time_of_day
        self._weather = weather
        self._road_type = road_type
        self._maneuver = maneuver
        self._direction = direction
        self._tag = tag

    def find_good_labeled_frames(
            self, trace_path: str) -> Tuple[np.ndarray, np.ndarray]:
        """ Find good frames based on video labels. Assume video labels have
            consistent timestampswith the specified master sensor/topic.

        Args:
            trace_path (str): path to a trace

        Returns:
            Return a tuple (``arr_1``, ``arr_2``), where ``arr_1`` is a boolean
            mask of good frames, and `arr_2` is timestamps of all good frames.
        """
        fpath = os.path.join(trace_path, 'video_labels.csv')
        has_video_label = os.path.exists(fpath)
        if has_video_label:
            is_good_frames = []
            good_timestamps = []
            with open(fpath, 'r') as f:
                reader = csv.DictReader(f,
                                        fieldnames=LabelSearch.FIELDS,
                                        delimiter=',')
                for line in reader:  # for each line
                    # assume consistency between master sensor and video labels
                    good_timestamps.append(float(line.pop('timestamp')))

                    match = True
                    for field in line.keys():
                        regex = getattr(self,
                                        '_' + field)  # get the search regex
                        res = re.search(regex, line[field])
                        if not res:
                            match = False
                            break
                    if match:
                        is_good_frames.append(True)
                    else:
                        is_good_frames.append(False)
            is_good_frames = np.array(is_good_frames)
            good_timestamps = np.array(good_timestamps)[is_good_frames]
        else:
            is_good_frames = None
            good_timestamps = None

        return is_good_frames, good_timestamps



import copy
import os
import glob
from typing import Dict, List, Any
import numpy as np
import h5py
from ffio import FFReader
import pyrender

from .camera_utils import CameraParams, ViewSynthesis, ZNEAR, ZFAR
from .BaseSensor import BaseSensor
from ..Entity import Entity
from ...utils import logging, misc, transform


class Camera(BaseSensor):
    """ A RGB camera sensor object that synthesizes RGB image locally around the
    dataset given a viewpoint (potentially different from the dataset) and timestamp.

    Args:
        attach_to (Entity): A parent object (Car) to be attached to.
        config (Dict): Configuration of the sensor. An example (default) is,

            >>> DEFAULT_CONFIG = {
                'depth_mode': 'FIXED_PLANE',
                'znear': ZNEAR,
                'zfar': ZFAR,
                'use_lighting': False,
                'directional_light_intensity': 10,
                'recoloring_factor': 0.5,
                'use_synthesizer': True,
            }

            Check :class:`Viewsynthesis` object for more details about the configuration.

    """
    DEFAULT_CONFIG = {
        'name': 'camera_front',
        'depth_mode': 'FIXED_PLANE',
        'znear': ZNEAR,
        'zfar': ZFAR,
        'use_lighting': False,
        'directional_light_intensity': 10,
        'recoloring_factor': 0.5,
        'use_synthesizer': True,
    }

    def __init__(self, attach_to: Entity, config: Dict) -> None:
        super(Camera, self).__init__(attach_to, config)

        # Define the CameraParams object for each physical camera in each trace
        self._input_cams = {
            trace: CameraParams(trace.param_file, self.name)
            for trace in attach_to.parent.traces
        }
        for _cp in self._input_cams.values():
            _cp.resize(*config['size'])

        # Now define the virtual camera in VISTA (defaults copied from the
        # first trace)
        first_trace = attach_to.parent.traces[0]
        self._virtual_cam = copy.deepcopy(self._input_cams[first_trace])
        self._virtual_cam.resize(*config['size'])

        self._streams: Dict[str, FFReader] = dict()
        self._flow_streams: Dict[str, List[FFReader]] = dict()
        self._flow_meta: Dict[str, h5py.File] = dict()
        if self._config.get('use_synthesizer', True):
            self._view_synthesis: ViewSynthesis = ViewSynthesis(
                self._virtual_cam, self._config)
        else:
            self._view_synthesis = None

    def reset(self) -> None:
        """ Reset RGB camera sensor by initiating RGB data stream based on
        current reference pointer to the dataset.

        """
        logging.info(f'Camera ({self.id}) reset')

        # Close stream already open
        for stream in self.streams.values():
            stream.close()

        for flow_stream in [
                _vv for _v in self.flow_streams.values()
                for _vv in _v.values()
        ]:
            flow_stream.close()

        # Fetch video stream from the associated trace. All video streams are handled by the main
        # camera and shared across all cameras in an agent.
        multi_sensor = self.parent.trace.multi_sensor
        trace_camera = self._input_cams[self.parent.trace]
        if self.name == multi_sensor.main_camera:
            for camera_name in multi_sensor.camera_names:
                # get video stream
                video_path = os.path.join(self.parent.trace.trace_path,
                                          camera_name + '.avi')
                cam_h, cam_w = (trace_camera.get_height(),
                                trace_camera.get_width())
                stream = FFReader(video_path,
                                  custom_size=(cam_h, cam_w),
                                  verbose=False)
                self._streams[camera_name] = stream

                # seek based on timestamp
                frame_num = self.parent.trace.good_frames[camera_name][
                    self.parent.segment_index][self.parent.frame_index]
                seek_sec = self._streams[camera_name].frame_to_secs(frame_num)
                self._streams[camera_name].seek(seek_sec)

                # get flow
                flow_meta_path = os.path.join(self.parent.trace.trace_path,
                                              camera_name + '_flow_meta.h5')
                if os.path.exists(flow_meta_path):
                    self._flow_meta[camera_name] = h5py.File(
                        flow_meta_path, 'r')

                    self._flow_streams[camera_name] = dict()
                    for flow_name in self.flow_meta[camera_name].keys():
                        flow_path = os.path.join(
                            self.parent.trace.trace_path,
                            camera_name + '_flow_{}.mp4'.format(flow_name))
                        flow_stream = FFReader(flow_path, verbose=False)
                        flow_frame_num = frame_num  # flow for (frame_num, frame_num + 1)
                        flow_seek_sec = flow_stream.frame_to_secs(
                            flow_frame_num)
                        flow_stream.seek(flow_seek_sec)
                        self._flow_streams[camera_name][
                            flow_name] = flow_stream
                else:
                    logging.warning('No flow data')
        else:  # use shared streams from the main camera
            main_name = multi_sensor.main_camera
            main_sensor = [
                _s for _s in self.parent.sensors if _s.name == main_name
            ]
            assert len(main_sensor) == 1, 'Cannot find main sensor {}'.format(
                main_name)
            main_sensor = main_sensor[0]
            assert isinstance(main_sensor,
                              Camera), 'Main sensor is not Camera object'
            self._streams = main_sensor.streams

            self._flow_streams = main_sensor.flow_streams
            self._flow_meta = main_sensor.flow_meta

        # Add background mesh for all video stream in view synthesis
        if self.view_synthesis is not None:
            parent_sensor_dict = {_s.name: _s for _s in self.parent.sensors}
            for camera_name in self.streams.keys():
                if camera_name not in self.view_synthesis.bg_mesh_names:
                    if camera_name in parent_sensor_dict.keys():
                        camera_param = parent_sensor_dict[
                            camera_name]._input_cams[self.parent.trace]
                    else:
                        camera_param = self._virtual_cam
                    self.view_synthesis.add_bg_mesh(camera_param)

    def capture(self, timestamp: float, **kwargs) -> np.ndarray:
        """ Synthesize RGB image based on current timestamp and transformation
        between the novel viewpoint to be simulated and the nominal viewpoint from
        the pre-collected dataset. Note that if there exists optical flow data in
        the trace directory, the :class:`Camera` object will take the optical flow
        to interpolate across frame to the exact timestamp as opposed to retrieving
        the RGB frame with the closest timestamp in the dataset.

        Args:
            timestamp (float): Timestamp that allows to retrieve a pointer to
                the dataset for data-driven simulation (synthesizing RGB image
                from real RGB video).

        Returns:
            np.ndarray: A synthesized RGB image.

        """
        logging.info(f'Camera ({self.id}) capture')

        # Get frame at the closest smaller timestamp from dataset
        multi_sensor = self.parent.trace.multi_sensor
        if self.name == multi_sensor.main_camera:
            fetch_smaller = self.flow_streams != dict(
            )  # only when using optical flow
            all_frame_nums = multi_sensor.get_frames_from_times([timestamp],
                                                                fetch_smaller)
            for camera_name in multi_sensor.camera_names:
                # rgb camera stream
                stream = self.streams[camera_name]
                frame_num = all_frame_nums[camera_name][0]

                if frame_num < stream.frame_num:
                    seek_sec = stream.frame_to_secs(frame_num)
                    stream.seek(seek_sec)

                while stream.frame_num != frame_num:
                    stream.read()

                # flow stream
                if self.flow_streams != dict():
                    flow_frame_num = frame_num  # flow for (frame_num, frame_num + 1)
                    for flow_stream in self.flow_streams[camera_name].values():
                        if flow_frame_num < flow_stream.frame_num:
                            flow_seek_sec = flow_stream.frame_to_secs(
                                flow_frame_num)
                            flow_stream.seek(flow_seek_sec)

                        while flow_stream.frame_num != flow_frame_num:
                            flow_stream.read()

        frames = dict()
        for camera_name in multi_sensor.camera_names:
            frames[camera_name] = self.streams[camera_name].image.copy()

        # Interpolate frame at the exact timestamp
        if self.flow_streams != dict():
            for camera_name in self.flow_streams.keys():
                frame = frames[camera_name]

                flow = dict()
                for flow_name, flow_minmax in self.flow_meta[
                        camera_name].items():
                    flow_stream = self.flow_streams[camera_name][flow_name]
                    flow[flow_name] = misc.img2flow(
                        flow_stream.image.copy(),
                        flow_minmax[int(flow_stream.frame_num)],
                        frame.shape[:2])

                frame_num = int(self.streams[camera_name].frame_num)
                curr_ref_ts = multi_sensor.get_time_from_frame_num(
                    camera_name, frame_num)
                next_ref_ts = multi_sensor.get_time_from_frame_num(
                    camera_name, frame_num + 1)

                logging.warning(
                    'Stream frame number exceed 1 non-intentionally')
                self.streams[camera_name].read(
                )  # NOTE: stream frame number exceed 1 here
                next_frame = self.streams[camera_name].image.copy()
                frames[camera_name] = misc.biinterp(frame, next_frame,
                                                    flow['forward'],
                                                    flow['backward'],
                                                    timestamp, curr_ref_ts,
                                                    next_ref_ts)

        # Synthesis by rendering
        if self.view_synthesis is not None:
            latlongyaw = self.parent.relative_state.numpy()
            trans, rot = transform.latlongyaw2vec(latlongyaw)
            rendered_frame, _ = self.view_synthesis.synthesize(
                trans, rot, frames)
        else:
            rendered_frame = frames[self.name]

        return rendered_frame

    def update_scene_object(self, name: str, scene_object: pyrender.Mesh,
                            pose: Any) -> None:
        """ Update pyrender mesh object in the scene for rendering.

        Args:
            name (str): Name of the scene object.
            scene_object (pyrender.Mesh): The scene object.
            pose (Any): The pose of the scene object.

        """
        trans, rotvec = transform.latlongyaw2vec(pose)
        quat = transform.euler2quat(rotvec)
        self.view_synthesis.update_object_node(name, scene_object, trans, quat)

    @property
    def config(self) -> Dict:
        """ Configuration of the RGB camera sensor. """
        return self._config

    @property
    def camera_param(self) -> CameraParams:
        """ Camera parameters of the virtual camera. """
        return self._virtual_cam

    @property
    def streams(self) -> Dict[str, FFReader]:
        """ Data stream of RGB image/video dataset to be simulated from. """
        return self._streams

    @property
    def flow_streams(self) -> Dict[str, List[FFReader]]:
        """ Data stream of optical flow (if any). """
        return self._flow_streams

    @property
    def flow_meta(self) -> Dict[str, h5py.File]:
        """ Meta data of optical flow (if any). """
        return self._flow_meta

    @property
    def view_synthesis(self) -> ViewSynthesis:
        """ View synthesizer object. """
        return self._view_synthesis

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} (id={self.id})> ' + \
               f'name: {self.name} ' + \
               f'size: {self.virtual_cam.get_height()}x{self.virtual_cam.get_width()} ' + \
               f'#streams: {len(self.streams)} '



from typing import Optional, List, Dict, Any
import numpy as np
from shapely.geometry import box as Box
from shapely import affinity

from .. import World
from ..entities.agents.Car import Car
from ..entities.agents.Dynamics import StateDynamics
from ..entities.sensors.MeshLib import MeshLib
from ..utils import logging, misc, transform


def default_terminal_condition(task, agent_id, **kwargs):
    """ An example definition of terminal condition. """

    agent = [_a for _a in task.world.agents if _a.id == agent_id][0]

    def _check_out_of_lane():
        road_half_width = agent.trace.road_width / 2.
        return np.abs(agent.relative_state.x) > road_half_width

    def _check_exceed_max_rot():
        maximal_rotation = np.pi / 10.
        return np.abs(agent.relative_state.yaw) > maximal_rotation

    def _check_crash():
        other_agents = [_a for _a in task.world.agents if _a.id != agent_id]
        agent2poly = lambda _x: misc.agent2poly(
            _x, ref_dynamics=agent.human_dynamics)
        poly = agent2poly(agent)
        other_polys = list(map(agent2poly, other_agents))
        overlap = compute_overlap(poly, other_polys) / poly.area
        crashed = np.any(overlap > task.config['overlap_threshold'])
        return crashed

    out_of_lane = _check_out_of_lane()
    exceed_max_rot = _check_exceed_max_rot()
    crashed = _check_crash()
    done = out_of_lane or exceed_max_rot or crashed or agent.done
    other_info = {
        'done': done,
        'out_of_lane': out_of_lane,
        'exceed_max_rot': exceed_max_rot,
        'crashed': crashed,
    }

    return done, other_info


def default_reward_fn(task, agent_id, **kwargs):
    """ An example definition of reward function. """
    reward = -1 if kwargs['done'] else 0  # simply encourage survival

    return reward, {}


class MultiAgentBase:
    """ This class builds a simple environment with multiple cars in the scene, which
    involves randomly initializing ado cars in the front of the ego car, checking collision
    between cars, handling meshes for all virtual agents, and determining terminal condition.

    Args:
        trace_paths (List[str]): A list of trace paths.
        trace_config (Dict): Configuration of the trace.
        car_configs (List[Dict]): Configuration of ``every`` cars.
        sensors_configs (List[Dict]): Configuration of sensors on ``every`` cars.
        task_config (Dict): Configuration of the task. An example (default) is,

            >>> DEFAULT_CONFIG = {
                    'n_agents': 1,
                    'mesh_dir': None,
                    'overlap_threshold': 0.05,
                    'max_resample_tries': 10,
                    'init_dist_range': [5., 10.],
                    'init_lat_noise_range': [-1., 1.],
                    'init_yaw_noise_range': [-0.0, 0.0],
                    'reward_fn': default_reward_fn,
                    'terminal_condition': default_terminal_condition
                }

            Note that both ``reward_fn`` and ``terminal_condition`` have function signature
            as ``f(task, agent_id, **kwargs) -> (value, dict)``. For more details, please check
            the source code.
        logging_level (str): Logging level (``DEBUG``, ``INFO``, ``WARNING``,
                             ``ERROR``, ``CRITICAL``); default set to ``WARNING``.

    """
    DEFAULT_CONFIG = {
        'n_agents': 1,
        'mesh_dir': None,
        'overlap_threshold': 0.05,
        'max_resample_tries': 10,
        'init_dist_range': [5., 10.],
        'init_lat_noise_range': [-1., 1.],
        'init_yaw_noise_range': [-0.0, 0.0],
        'reward_fn': default_reward_fn,
        'terminal_condition': default_terminal_condition
    }

    def __init__(self,
                 trace_paths: List[str],
                 trace_config: Dict,
                 car_configs: List[Dict],
                 sensors_configs: List[List[Dict]],
                 task_config: Optional[Dict] = dict(),
                 logging_level: Optional[str] = 'WARNING'):
        logging.setLevel(getattr(logging, logging_level))
        self._config = misc.merge_dict(task_config, self.DEFAULT_CONFIG)
        n_agents = self.config['n_agents']
        assert len(
            car_configs
        ) == n_agents, 'Number of car config is not consistent with number of agents'
        assert len(
            sensors_configs
        ) == n_agents, 'Number of sensors config is not consistent with number of agents'
        assert car_configs[0][
            'lookahead_road'], '\'lookahead_road\' in the first car config should be set to True'

        self._world: World = World(trace_paths, trace_config)
        for i in range(n_agents):
            agent = self._world.spawn_agent(car_configs[i])
            for sensor_config in sensors_configs[i]:
                sensor_type = sensor_config.pop('type')
                if sensor_type == 'camera':
                    agent.spawn_camera(sensor_config)
                else:
                    raise NotImplementedError(
                        f'Unrecognized sensor type {sensor_type}')

        if n_agents > 1:
            assert self.config[
                'mesh_dir'] is not None, 'Specify mesh_dir if n_agents > 1'
            self._meshlib = MeshLib(self.config['mesh_dir'])

        self.set_seed(0)

    def reset(self) -> Dict:
        """ Reset the environment. This involves regular world reset, randomly
        initializing ado agent in the front of the ego agent, and resetting
        the mesh library for all virtual agents.

        Returns:
            Dict: A dictionary with keys as agent IDs and values as observation
            for each agent, which is also a dictionary with keys as sensor IDs
            and values as sensory measurements.

        """
        # Reset world; all agents are initialized at the same pointer to the trace
        new_trace_index, new_segment_index, new_frame_index = \
            self.world.sample_new_location()
        for agent in self.world.agents:
            agent.reset(new_trace_index,
                        new_segment_index,
                        new_frame_index,
                        step_sensors=False)

        # Randomly initialize ado agents in the front
        ref_dynamics = self.ego_agent.human_dynamics
        polys = [misc.agent2poly(self.ego_agent, ref_dynamics)]
        for agent in self.world.agents:
            if agent == self.ego_agent:
                continue

            collision_free = False
            resample_tries = 0
            while not collision_free and resample_tries < self.config[
                    'max_resample_tries']:
                self._randomly_place_agent(agent)
                poly = misc.agent2poly(agent, ref_dynamics)
                overlap = compute_overlap(poly, polys) / poly.area
                collision_free = np.all(
                    overlap <= self.config['overlap_threshold'])

                resample_tries += 1
            polys.append(poly)

        # Reset mesh library
        if len(self.world.agents) > 1:
            self._reset_meshlib()

        # Get observation
        self._sensor_capture()
        observations = {_a.id: _a.observations for _a in self.world.agents}

        return observations

    def step(self, actions, dt=1 / 30.):
        """ Step the environment. This includes updating agents' states, synthesizing
        agents' observations, checking terminal conditions, and computing rewards.

        Args:
            actions (Dict[str, np.ndarray]):
                A dictionary with keys as agent IDs and values as actions
                to be executed to interact with the environment and other
                agents.
            dt (float): Elapsed time in second; default set to 1/30.

        Returns:
            Return a tuple (``dict_a``, ``dict_b``, ``dict_c``, ``dict_d``),
            where ``dict_a`` is the observation, ``dict_b`` is the reward,
            ``dict_c`` is whether the episode terminates, ``dict_d`` is additional
            informations for every agents; keys of every dictionary are agent IDs.

        """
        # Update agents' dynamics (state)
        for agent in self.world.agents:
            action = actions[agent.id]
            agent.step_dynamics(action, dt=dt)

        # Get agents' sensory measurement
        self._sensor_capture()
        observations = {_a.id: _a.observations for _a in self.world.agents}

        # Check terminal conditions
        dones = dict()
        infos_from_terminal_condition = dict()
        terminal_condition = self.config['terminal_condition']
        for agent in self.world.agents:
            dones[agent.id], infos_from_terminal_condition[
                agent.id] = terminal_condition(self, agent.id)

        # Compute reward
        rewards = dict()
        reward_fn = self.config['reward_fn']
        for agent in self.world.agents:
            rewards[agent.id], _ = reward_fn(
                self, agent.id, **infos_from_terminal_condition[agent.id])

        # Get info
        infos = dict()

        return observations, rewards, dones, infos

    def set_seed(self, seed) -> None:
        """ Set random seed.

        Args:
            seed (int): Random seed.

        """
        self._seed = seed
        self._rng = np.random.default_rng(self.seed)
        self.world.set_seed(seed)

    def _randomly_place_agent(self, agent: Car):
        # Randomly sampled a pose in the front of ego agent that is still on
        # the road. This can be achieved by,
        # (1) randomly sampling a distance from the ego agent
        # (2) fetch the closest pointer from the road kept by the ego agent
        # (3) slightly perturb the associated pose.
        tgt_dist = self._rng.uniform(*self.config['init_dist_range'])

        road = np.array(self.ego_agent.road)
        dist_from_ego = np.linalg.norm(road[:, :2], axis=1)
        tgt_idx = np.argmin(np.abs(tgt_dist - dist_from_ego))
        tgt_pose = road[tgt_idx].copy()

        lat_noise = self._rng.uniform(*self.config['init_lat_noise_range'])
        tgt_pose[0] += lat_noise * np.cos(tgt_pose[2])
        tgt_pose[1] += lat_noise * np.sin(tgt_pose[2])
        yaw_noise = self._rng.uniform(*self.config['init_yaw_noise_range'])
        tgt_pose[2] += yaw_noise

        # Place agent given the randomly sampled pose
        agent.ego_dynamics.update(*tgt_pose)
        agent.step_dynamics(tgt_pose[-2:], dt=1e-8)

    def _reset_meshlib(self):
        self._meshlib.reset(self.config['n_agents'])

        # Assign car width and length based on mesh size
        for i, agent in enumerate(self.world.agents):
            agent._width = self._meshlib.agents_meshes_dim[i][0]
            agent._length = self._meshlib.agents_meshes_dim[i][1]

    def _sensor_capture(self) -> None:
        # Update mesh of virtual agents
        if self.config['n_agents'] > 1:
            for agent in self.world.agents:
                if len(agent.sensors) == 0:
                    continue

                for i, other_agent in enumerate(self.world.agents):
                    if other_agent.id == agent.id:
                        continue
                    # compute relative pose to the ego agent
                    latlongyaw = transform.compute_relative_latlongyaw(
                        other_agent.ego_dynamics.numpy()[:3],
                        agent.human_dynamics.numpy()[:3])
                    # add to sensor
                    scene_object = self._meshlib.agents_meshes[i]
                    name = f'agent_{i}'
                    for sensor in agent.sensors:
                        sensor.update_scene_object(name, scene_object,
                                                   latlongyaw)

        # Step sensors
        for agent in self.world.agents:
            agent.step_sensors()

    @property
    def config(self) -> Dict:
        """ Configuration of this task. """
        return self._config

    @property
    def ego_agent(self) -> Car:
        """ Ego agent. """
        return self.world.agents[0]

    @property
    def world(self) -> World:
        """ :class:`World` of this task. """
        return self._world

    @property
    def seed(self) -> int:
        """ Random seed for the task and the associated :class:`World`. """
        return self._seed


def compute_overlap(poly: Box, polys: List[Box]) -> List[float]:
    """ Compute overlapping area between 1 polygons and N polygons.

    Args:
        poly (shapely.geometry.Box): A polygon.
        poly (List): A list of polygon.

    Returns:
        List[float]: Intersecting area between polygons.

    """
    n_polys = len(polys)
    overlap = np.zeros((n_polys))
    for i in range(n_polys):
        intersection = polys[i].intersection(poly)
        overlap[i] = intersection.area
    return overlap
